@@ -24,22 +24,30 @@ be specific to loom, to ensure people have the loom plugin when working on a
 loom branch.
 """
 
+from __future__ import absolute_import
+
 from StringIO import StringIO
 
 import bzrlib.branch
-from bzrlib import bzrdir
+from bzrlib import (
+    bzrdir,
+    errors,
+    fetch as _mod_fetch,
+    remote,
+    symbol_versioning,
+    trace,
+    tree as _mod_tree,
+    ui,
+    urlutils,
+    )
 from bzrlib.decorators import needs_read_lock, needs_write_lock
-import bzrlib.errors
-import bzrlib.osutils
-from bzrlib import symbol_versioning
-import bzrlib.trace
-import bzrlib.ui
 from bzrlib.revision import is_null, NULL_REVISION
-import bzrlib.tree
-import bzrlib.urlutils
 
-import loom_io
-import loom_state
+from bzrlib.plugins.loom import (
+    formats,
+    loom_io,
+    loom_state,
+    )
 
 
 EMPTY_REVISION = 'empty:'
@@ -51,11 +59,20 @@ def create_thread(loom, thread_name):
     loom.lock_write()
     try:
         loom.new_thread(thread_name, loom.nick)
-        loom.nick = thread_name
+        loom._set_nick(thread_name)
     finally:
         loom.unlock()
 
 
+class AlreadyLoom(errors.BzrError):
+
+    _fmt = """Loom %(loom)s is already a loom."""
+
+    def __init__(self, loom):
+        errors.BzrError.__init__(self)
+        self.loom = loom
+
+
 def loomify(branch):
     """Convert branch to a loom.
 
@@ -63,6 +80,12 @@ def loomify(branch):
     """
     try:
         branch.lock_write()
+        try:
+            require_loom_branch(branch)
+        except NotALoom:
+            pass
+        else:
+            raise AlreadyLoom(branch)
         try:
             format = {
                 bzrlib.branch.BzrBranchFormat5: BzrBranchLoomFormat1,
@@ -76,43 +99,30 @@ def loomify(branch):
         branch.unlock()
 
 
-def require_loom_branch(branch):
-    """Return None if branch is already loomified, or raise NotALoom."""
-    if not branch._format.__class__ in LOOM_FORMATS:
-        raise NotALoom(branch)
+require_loom_branch = formats.require_loom_branch
+NotALoom = formats.NotALoom
 
 
-class NotALoom(bzrlib.errors.BzrError):
-
-    _fmt = ("The branch %(branch)s is not a loom. "
-        "You can use 'bzr loomify' to make it into a loom.")
-
-    def __init__(self, branch):
-        bzrlib.errors.BzrError.__init__(self)
-        self.branch = branch
-
-
-class LoomThreadError(bzrlib.errors.BzrError):
-
-    _fmt = """Base class for Loom-Thread errors."""
+class LoomThreadError(errors.BzrError):
+    """Base class for Loom-Thread errors."""
 
     def __init__(self, branch, thread):
-        bzrlib.errors.BzrError.__init__(self)
+        errors.BzrError.__init__(self)
         self.branch = branch
         self.thread = thread
 
 
-class UnrecordedRevision(bzrlib.errors.BzrError):
+class UnrecordedRevision(errors.BzrError):
 
     _fmt = """The revision %(revision_id)s is not recorded in the loom %(branch)s."""
 
     def __init__(self, branch, revision_id):
-        bzrlib.errors.BzrError.__init__(self)
+        errors.BzrError.__init__(self)
         self.branch = branch
         self.revision_id = revision_id
 
 
-class UnsupportedBranchFormat(bzrlib.errors.BzrError):
+class UnsupportedBranchFormat(errors.BzrError):
 
     _fmt = """The branch format %(format)s is not supported by loomify."""
 
@@ -135,7 +145,7 @@ class NoSuchThread(LoomThreadError):
     _fmt = """No such thread '%(thread)s'."""
 
 
-class NoLowerThread(bzrlib.errors.BzrError):
+class NoLowerThread(errors.BzrError):
 
     _fmt = """No lower thread exists."""
 
@@ -145,7 +155,7 @@ class CannotCombineOnLastThread(NoLowerThread):
     _fmt = """Cannot combine threads on the bottom thread."""
 
 
-class LoomMetaTree(bzrlib.tree.Tree):
+class LoomMetaTree(_mod_tree.InventoryTree):
     """A 'tree' object that is used to commit the loom meta branch."""
 
     def __init__(self, loom_meta_ie, loom_stream, loom_sha1):
@@ -209,14 +219,14 @@ class LoomSupport(object):
             if len(threads) <= current_index:
                 # removed the end
                 # take the new end thread
-                self.nick = threads[-1][0]
+                self._set_nick(threads[-1][0])
                 new_rev = threads[-1][1]
                 if new_rev == EMPTY_REVISION:
                     new_rev = bzrlib.revision.NULL_REVISION
                 self.generate_revision_history(new_rev)
                 return
             # non-end thread removed.
-            self.nick = threads[current_index][0]
+            self._set_nick(threads[current_index][0])
             new_rev = threads[current_index][1]
             if new_rev == EMPTY_REVISION:
                 new_rev = bzrlib.revision.NULL_REVISION
@@ -234,14 +244,16 @@ class LoomSupport(object):
         :type other: Branch
         """
         # Looms are not currently bindable.
-        raise bzrlib.errors.UpgradeRequired(self.base)
+        raise errors.UpgradeRequired(self.base)
 
     @needs_read_lock
-    def clone(self, to_bzrdir, revision_id=None, repository_policy=None):
+    def clone(self, to_bzrdir, revision_id=None, repository_policy=None, name=None):
         """Clone the branch into to_bzrdir.
         
-        This differs from the base clone by cloning the loom and 
-        setting the current nick to the top of the loom.
+        This differs from the base clone by cloning the loom, setting the
+        current nick to the top of the loom, not honouring any branch format
+        selection on the target bzrdir, and ensuring that the format of
+        the created branch is stacking compatible.
         """
         # If the target is a stackable repository, force-upgrade the
         # output loom format
@@ -250,83 +262,14 @@ class LoomSupport(object):
             format = BzrBranchLoomFormat7()
         else:
             format = self._format
-        result = format.initialize(to_bzrdir)
+        result = format.initialize(to_bzrdir, name=name)
         if repository_policy is not None:
             repository_policy.configure_branch(result)
-        self.copy_content_into(result, revision_id=revision_id)
+        bzrlib.branch.InterBranch.get(self, result).copy_content_into(
+            revision_id=revision_id)
         return result
 
-    @needs_read_lock
-    def copy_content_into(self, destination, revision_id=None):
-        # XXX: hint for bzrlib - break this into two routines, one for
-        # copying the last-rev pointer, one for copying parent etc.
-        destination.lock_write()
-        try:
-            source_nick = self.nick
-            state = self.get_loom_state()
-            parents = state.get_parents()
-            if parents:
-                loom_tip = parents[0]
-            else:
-                loom_tip = None
-            threads = self.get_threads(state.get_basis_revision_id())
-            if revision_id not in (None, NULL_REVISION):
-                if threads:
-                    # revision_id should be in the loom, or its an error 
-                    found_threads = [thread for thread, rev in threads
-                        if rev == revision_id]
-                    if not found_threads:
-                        # the thread we have been asked to set in the remote 
-                        # side has not been recorded yet, so its data is not
-                        # present at this point.
-                        raise UnrecordedRevision(self, revision_id)
-
-                # pull in the warp, which was skipped during the initial pull
-                # because the front end does not know what to pull.
-                # nb: this is mega huge hacky. THINK. RBC 2006062
-                nested = bzrlib.ui.ui_factory.nested_progress_bar()
-                try:
-                    if parents:
-                        destination.repository.fetch(self.repository,
-                            revision_id=parents[0])
-                    if threads:
-                        for thread, rev_id in reversed(threads):
-                            # fetch the loom content for this revision
-                            destination.repository.fetch(self.repository,
-                                revision_id=rev_id)
-                finally:
-                    nested.finished()
-            state = loom_state.LoomState()
-            try:
-                require_loom_branch(destination)
-                if threads:
-                    last_rev = threads[-1][1]
-                    if last_rev == EMPTY_REVISION:
-                        last_rev = bzrlib.revision.NULL_REVISION
-                    destination.generate_revision_history(last_rev)
-                    state.set_parents([loom_tip])
-                    state.set_threads(
-                        (thread + ([thread[1]],) for thread in threads)
-                        )
-                else:
-                    # no threads yet, be a normal branch.
-                    self._synchronize_history(destination, revision_id)
-                destination._set_last_loom(state)
-            except NotALoom:
-                self._synchronize_history(destination, revision_id)
-            try:
-                parent = self.get_parent()
-            except bzrlib.errors.InaccessibleParent, e:
-                bzrlib.trace.mutter('parent was not accessible to copy: %s', e)
-            else:
-                if parent:
-                    destination.set_parent(parent)
-            if threads:
-                destination.nick = threads[-1][0]
-        finally:
-            destination.unlock()
-
-    def _get_checkout_format(self):
+    def _get_checkout_format(self, lightweight=False):
         """Checking out a Loom gets a regular branch for now.
         
         This is a short term measure to get to an all-tests passing status.
@@ -346,7 +289,7 @@ class LoomSupport(object):
     def get_old_bound_location(self):
         """Return the URL of the branch we used to be bound to."""
         # No binding for looms yet.
-        raise bzrlib.errors.UpgradeRequired(self.base)
+        raise errors.UpgradeRequired(self.base)
 
     def get_threads(self, rev_id):
         """Return the threads from a loom revision.
@@ -372,14 +315,14 @@ class LoomSupport(object):
         threads = self.get_loom_state().get_threads()
         for thread_name, thread_revision, _parents in threads:
             thread_transport = root_transport.clone(thread_name)
-            user_location = bzrlib.urlutils.unescape_for_display(
+            user_location = urlutils.unescape_for_display(
                 thread_transport.base, 'utf-8')
             try:
                 control_dir = bzrdir.BzrDir.open(thread_transport.base,
                     possible_transports=[thread_transport])
                 tree, branch = control_dir._get_tree_branch()
-            except bzrlib.errors.NotBranchError:
-                bzrlib.trace.note('Creating branch at %s' % user_location)
+            except errors.NotBranchError:
+                trace.note('Creating branch at %s' % user_location)
                 branch = bzrdir.BzrDir.create_branch_convenience(
                     thread_transport.base,
                     possible_transports=[thread_transport])
@@ -387,11 +330,11 @@ class LoomSupport(object):
                     thread_transport.base)
             else:
                 if thread_revision == branch.last_revision():
-                    bzrlib.trace.note('Skipping up-to-date branch at %s'
+                    trace.note('Skipping up-to-date branch at %s'
                                       % user_location)
                     continue
                 else:
-                    bzrlib.trace.note('Updating branch at %s' % user_location)
+                    trace.note('Updating branch at %s' % user_location)
             if tree is not None:
                 tree.pull(self, stop_revision=thread_revision)
             else:
@@ -453,32 +396,48 @@ class LoomSupport(object):
             result.append((name, rev_id))
         return result
 
-    @needs_write_lock
-    def pull(self, source, overwrite=False, stop_revision=None,
-        run_hooks=True, possible_transports=None, _override_hook_target=None,
-        local=False):
-        """Pull from a branch into this loom.
+    def _loom_get_nick(self):
+        return self._get_nick(local=True)
 
-        If the remote branch is a non-loom branch, the pull is done against the
-        current warp. If it is a loom branch, then the pull is done against the
-        entire loom and the current thread set to the top thread.
-        """
-        if not isinstance(source, LoomSupport):
-            return super(LoomSupport, self).pull(source,
-                overwrite=overwrite, stop_revision=stop_revision,
-                possible_transports=possible_transports,
-                _override_hook_target=_override_hook_target, local=local)
-        return _Puller(source, self).transfer(overwrite, stop_revision,
-            run_hooks, possible_transports, _override_hook_target, local)
+    def _rename_thread(self, nick):
+        """Rename the current thread to nick."""
+        state = self.get_loom_state()
+        threads = state.get_threads()
+        if not len(threads):
+            # No threads at all - probably a default initialised loom in the
+            # test suite.
+            return  self._set_nick(nick)
+        current_index = state.thread_index(self.nick)
+        threads[current_index] = (nick,) + threads[current_index][1:]
+        state.set_threads(threads)
+        self._set_last_loom(state)
+        # Preserve default behavior: set the branch nick
+        self._set_nick(nick)
+
+    nick = property(_loom_get_nick, _rename_thread)
+
+    def heads_to_fetch(self):
+        """See Branch.heads_to_fetch."""
+        # The base implementation returns ([tip], tags)
+        must_fetch, should_fetch = super(LoomSupport, self).heads_to_fetch()
+        # Add each thread's content to must_fetch
+        must_fetch.update(
+            thread_rev for thread_name, thread_rev, thread_parents in
+            self.get_loom_state().get_threads())
+        must_fetch.discard(EMPTY_REVISION)
+        must_fetch.discard(bzrlib.revision.NULL_REVISION)
+        return must_fetch, should_fetch
 
     @needs_read_lock
-    def push(self, target, overwrite=False, stop_revision=None,
+    def push(self, target, overwrite=False, stop_revision=None, lossy=False,
         _override_hook_source_branch=None):
         # Not ideal, but see the issues raised on bazaar@lists.canonical.com
         # about the push api needing work.
         if not isinstance(target, LoomSupport):
             return super(LoomSupport, self).push(target, overwrite,
-                stop_revision, _override_hook_source_branch=None)
+                stop_revision, lossy=lossy, _override_hook_source_branch=None)
+        if lossy:
+            raise errors.LossyPushToSameVCS(self, target)
         return _Pusher(self, target).transfer(overwrite, stop_revision,
                                               run_hooks=True)
 
@@ -494,7 +453,7 @@ class LoomSupport(object):
         threads = state.get_threads()
         # check the semantic value, not the serialised value for equality.
         if old_threads == threads:
-            raise bzrlib.errors.PointlessCommit
+            raise errors.PointlessCommit
         builder = self.get_commit_builder(parents)
         loom_ie = bzrlib.inventory.make_entry(
             'file', 'loom', bzrlib.inventory.ROOT_ID, 'loom_meta_tree')
@@ -510,7 +469,7 @@ class LoomSupport(object):
             builder.record_entry_contents(root_ie, [], '', loom_tree,
                 ('directory', None, None, None))
         builder.record_entry_contents(
-            loom_ie, parents, 'loom', loom_tree,
+            loom_ie, list(self.repository.iter_inventories(parents)), 'loom', loom_tree,
             # a fake contents so that the file is determined as changed.
             ('file', 0, False, None))
         builder.finish_inventory()
@@ -625,10 +584,20 @@ class LoomSupport(object):
 
 
 class _Puller(object):
+    # XXX: Move into InterLoomBranch.
 
     def __init__(self, source, target):
         self.target = target
         self.source = source
+        # If _Puller has been created, we need real branch objects.
+        self.real_target = self.unwrap_branch(target)
+        self.real_source = self.unwrap_branch(source)
+
+    def unwrap_branch(self, branch):
+        if isinstance(branch, remote.RemoteBranch):
+            branch._ensure_real()
+            return branch._real_branch
+        return branch
 
     def prepare_result(self, _override_hook_target):
         result = self.make_result()
@@ -643,7 +612,6 @@ class _Puller(object):
         return result
 
     def finish_result(self, result):
-        result.tag_conflicts = None
         result.new_revno, result.new_revid = self.target.last_revision_info()
 
     def do_hooks(self, result, run_hooks):
@@ -670,36 +638,42 @@ class _Puller(object):
             new_rev = self.source.last_revision()
         if new_rev == EMPTY_REVISION:
             new_rev = bzrlib.revision.NULL_REVISION
+        fetch_spec = self.build_fetch_spec(stop_revision)
         self.target.repository.fetch(self.source.repository,
-            revision_id=new_rev)
-        if not overwrite:
-            new_rev_ancestry = self.source.repository.get_ancestry(
-                new_rev)
-            last_rev = self.target.last_revision()
-            # get_ancestry returns None for NULL_REVISION currently.
-            if last_rev == NULL_REVISION:
-                last_rev = None
-            if last_rev not in new_rev_ancestry:
-                raise bzrlib.errors.DivergedBranches(
-                    self.target, self.source)
-        self.target.generate_revision_history(new_rev)
+            fetch_spec=fetch_spec)
+        self.target.generate_revision_history(new_rev, self.target.last_revision(),
+            self.source)
+        tag_ret = self.source.tags.merge_to(self.target.tags)
+        if isinstance(tag_ret, tuple):
+            result.tag_updates, result.tag_conflicts = tag_ret
+        else:
+            result.tag_conflicts = tag_ret
         # get the final result object details
         self.do_hooks(result, run_hooks)
         return result
 
+    def build_fetch_spec(self, stop_revision):
+        factory = _mod_fetch.FetchSpecFactory()
+        factory.source_branch = self.source
+        factory.source_repo = self.source.repository
+        factory.source_branch_stop_revision_id = stop_revision
+        factory.target_repo = self.target.repository
+        factory.target_repo_kind = _mod_fetch.TargetRepoKinds.PREEXISTING
+        return factory.make_fetch_spec()
+
     def transfer(self, overwrite, stop_revision, run_hooks=True,
         possible_transports=None, _override_hook_target=None, local=False):
         """Implementation of push and pull"""
         if local:
-            raise bzrlib.errors.LocalRequiresBoundBranch()
+            raise errors.LocalRequiresBoundBranch()
         # pull the loom, and position our
-        pb = bzrlib.ui.ui_factory.nested_progress_bar()
+        pb = ui.ui_factory.nested_progress_bar()
         try:
             result = self.prepare_result(_override_hook_target)
             self.target.lock_write()
             self.source.lock_read()
             try:
-                source_state = self.source.get_loom_state()
+                source_state = self.real_source.get_loom_state()
                 source_parents = source_state.get_parents()
                 if not source_parents:
                     return self.plain_transfer(result, run_hooks,
@@ -711,10 +685,10 @@ class _Puller(object):
                 if not overwrite:
                     # is the loom compatible?
                     if len(my_state.get_parents()) > 0:
-                        source_ancestry = self.source.repository.get_ancestry(
-                            source_loom_rev)
-                        if my_state.get_parents()[0] not in source_ancestry:
-                            raise bzrlib.errors.DivergedBranches(
+                        graph = self.source.repository.get_graph()
+                        if not graph.is_ancestor(my_state.get_parents()[0],
+                                source_loom_rev):
+                            raise errors.DivergedBranches(
                                 self.target, self.source)
                 # fetch the loom content
                 self.target.repository.fetch(self.source.repository,
@@ -724,17 +698,10 @@ class _Puller(object):
                     source_state.get_basis_revision_id())
                 # stopping at from our repository.
                 revisions = [rev for name,rev in threads]
-                # for each thread from top to bottom, retrieve its referenced
-                # content. XXX FIXME: a revision_ids parameter to fetch would be
-                # nice here.
-                # the order is reversed because the common case is for the top
-                # thread to include all content.
-                for rev_id in reversed(revisions):
-                    if rev_id not in (EMPTY_REVISION,
-                        bzrlib.revision.NULL_REVISION):
-                        # fetch the loom content for this revision
-                        self.target.repository.fetch(self.source.repository,
-                            revision_id=rev_id)
+                # fetch content for all threads and tags.
+                fetch_spec = self.build_fetch_spec(stop_revision)
+                self.target.repository.fetch(self.source.repository,
+                    fetch_spec=fetch_spec)
                 # set our work threads to match (this is where we lose data if
                 # there are local mods)
                 my_state.set_threads(
@@ -745,12 +712,18 @@ class _Puller(object):
                 # and save the state.
                 self.target._set_last_loom(my_state)
                 # set the branch nick.
-                self.target.nick = threads[-1][0]
+                self.target._set_nick(threads[-1][0])
                 # and position the branch on the top loom
                 new_rev = threads[-1][1]
                 if new_rev == EMPTY_REVISION:
                     new_rev = bzrlib.revision.NULL_REVISION
                 self.target.generate_revision_history(new_rev)
+                # merge tags
+                tag_ret = self.source.tags.merge_to(self.target.tags)
+                if isinstance(tag_ret, tuple):
+                    result.tag_updates, tag_conflicts = tag_ret
+                else:
+                    result.tag_conflicts = tag_ret
                 self.do_hooks(result, run_hooks)
                 return result
             finally:
@@ -801,11 +774,13 @@ class LoomFormatMixin(object):
     # A mixin is not ideal because it is tricky to test, but it seems to be the
     # best solution for now.
 
-    def initialize(self, a_bzrdir, name=None):
+    def initialize(self, a_bzrdir, name=None, repository=None,
+                   append_revisions_only=None):
         """Create a branch of this format in a_bzrdir."""
-        if name is not None:
-            raise bzrlib.errors.NoColocatedBranchSupport(self)
-        super(LoomFormatMixin, self).initialize(a_bzrdir, name=None)
+        super(LoomFormatMixin, self).initialize(a_bzrdir, name=name,
+                repository=repository,
+                append_revisions_only=append_revisions_only)
+
         branch_transport = a_bzrdir.get_branch_transport(self)
         files = []
         state = loom_state.LoomState()
@@ -818,35 +793,37 @@ class LoomFormatMixin(object):
             branch_transport, 'lock', bzrlib.lockdir.LockDir)
         control_files.lock_write()
         try:
-            for name, stream in files:
-                branch_transport.put_file(name, stream)
+            for filename, stream in files:
+                branch_transport.put_file(filename, stream)
         finally:
             control_files.unlock()
-        return self.open(a_bzrdir, _found=True, )
+        return self.open(a_bzrdir, _found=True, name=name)
 
-    def open(self, a_bzrdir, name=None, _found=False, ignore_fallbacks=False):
+    def open(self, a_bzrdir, name=None, _found=False, ignore_fallbacks=False,
+            found_repository=None, possible_transports=None):
         """Return the branch object for a_bzrdir
 
         _found is a private parameter, do not use it. It is used to indicate
                if format probing has already be done.
 
         :param name: The 'colocated branches' name for the branch to open.
-            in future, Loom may use that to return a Thread, but for now
-            it is unused.
         """
+        if name is None:
+            name = a_bzrdir._get_selected_branch()
         if not _found:
-            format = BranchFormat.find_format(a_bzrdir)
+            format = bzrlib.branch.BranchFormat.find_format(a_bzrdir, name=name)
             assert format.__class__ == self.__class__
-        if name is not None:
-            raise bzrlib.errors.NoColocatedBranchSupport(self)
-        transport = a_bzrdir.get_branch_transport(None)
+        transport = a_bzrdir.get_branch_transport(None, name=name)
         control_files = bzrlib.lockable_files.LockableFiles(
             transport, 'lock', bzrlib.lockdir.LockDir)
-        return self._branch_class(_format=self,
+        if found_repository is None:
+            found_repository = a_bzrdir.find_repository()
+        return self._branch_class()(_format=self,
                           _control_files=control_files,
                           a_bzrdir=a_bzrdir,
-                          _repository=a_bzrdir.find_repository(),
-                          ignore_fallbacks=ignore_fallbacks)
+                          _repository=found_repository,
+                          ignore_fallbacks=ignore_fallbacks,
+                          name=name)
 
     def take_over(self, branch):
         """Take an existing bzrlib branch over into Loom format.
@@ -879,10 +856,13 @@ class BzrBranchLoomFormat1(LoomFormatMixin, bzrlib.branch.BzrBranchFormat5):
     This format is new in the loom plugin.
     """
 
-    _branch_class = LoomBranch
+    def _branch_class(self):
+        return LoomBranch
+
     _parent_classs = bzrlib.branch.BzrBranchFormat5
 
-    def get_format_string(self):
+    @classmethod
+    def get_format_string(cls):
         """See BranchFormat.get_format_string()."""
         return "Bazaar-NG Loom branch format 1\n"
 
@@ -906,10 +886,13 @@ class BzrBranchLoomFormat6(LoomFormatMixin, bzrlib.branch.BzrBranchFormat6):
     This format is new in the loom plugin.
     """
 
-    _branch_class = LoomBranch6
+    def _branch_class(self):
+        return LoomBranch6
+
     _parent_classs = bzrlib.branch.BzrBranchFormat6
 
-    def get_format_string(self):
+    @classmethod
+    def get_format_string(cls):
         """See BranchFormat.get_format_string()."""
         return "Bazaar-NG Loom branch format 6\n"
 
@@ -933,10 +916,13 @@ class BzrBranchLoomFormat7(LoomFormatMixin, bzrlib.branch.BzrBranchFormat7):
     This format is new in the loom plugin.
     """
 
-    _branch_class = LoomBranch7
+    def _branch_class(self):
+        return LoomBranch7
+
     _parent_classs = bzrlib.branch.BzrBranchFormat7
 
-    def get_format_string(self):
+    @classmethod
+    def get_format_string(cls):
         """See BranchFormat.get_format_string()."""
         return "Bazaar-NG Loom branch format 7\n"
 
@@ -948,13 +934,137 @@ class BzrBranchLoomFormat7(LoomFormatMixin, bzrlib.branch.BzrBranchFormat7):
         return "bzr loom format 7 (based on bzr branch format 7)\n"
 
 
-bzrlib.branch.BranchFormat.register_format(BzrBranchLoomFormat1())
-bzrlib.branch.BranchFormat.register_format(BzrBranchLoomFormat6())
-bzrlib.branch.BranchFormat.register_format(BzrBranchLoomFormat7())
+# Handle the smart server:
+
+class InterLoomBranch(bzrlib.branch.GenericInterBranch):
+
+    @classmethod
+    def _get_branch_formats_to_test(klass):
+        default_format = bzrlib.branch.format_registry.get_default()
+        return [
+            (default_format, BzrBranchLoomFormat7()),
+            (BzrBranchLoomFormat7(), default_format),
+            (BzrBranchLoomFormat7(), BzrBranchLoomFormat7()),
+            ]
+
+    def unwrap_branch(self, branch):
+        if isinstance(branch, remote.RemoteBranch):
+            branch._ensure_real()
+            return branch._real_branch
+        return branch
+
+    @classmethod
+    def is_compatible(klass, source, target):
+        # 1st cut: special case and handle all *->Loom and Loom->*
+        return klass.branch_is_loom(source) or klass.branch_is_loom(target)
+
+    def get_loom_state(self, branch):
+        branch = self.unwrap_branch(branch)
+        return branch.get_loom_state()
+
+    def get_threads(self, branch, revision_id):
+        branch = self.unwrap_branch(branch)
+        return branch.get_threads(revision_id)
+
+    @classmethod
+    def branch_is_loom(klass, branch):
+        format = klass.unwrap_format(branch._format)
+        return isinstance(format, LoomFormatMixin)
+
+    @needs_write_lock
+    def copy_content_into(self, revision_id=None):
+        if not self.__class__.branch_is_loom(self.source):
+            # target is loom, but the generic code path works Just Fine for
+            # regular to loom copy_content_into.
+            return super(InterLoomBranch, self).copy_content_into(
+                revision_id=revision_id)
+        # XXX: hint for bzrlib - break this into two routines, one for
+        # copying the last-rev pointer, one for copying parent etc.
+        source_nick = self.source.nick
+        state = self.get_loom_state(self.source)
+        parents = state.get_parents()
+        if parents:
+            loom_tip = parents[0]
+        else:
+            loom_tip = None
+        threads = self.get_threads(self.source, state.get_basis_revision_id())
+        if revision_id not in (None, NULL_REVISION):
+            if threads:
+                # revision_id should be in the loom, or its an error 
+                found_threads = [thread for thread, rev in threads
+                    if rev == revision_id]
+                if not found_threads:
+                    # the thread we have been asked to set in the remote 
+                    # side has not been recorded yet, so its data is not
+                    # present at this point.
+                    raise UnrecordedRevision(self.source, revision_id)
+
+            # pull in the warp, which was skipped during the initial pull
+            # because the front end does not know what to pull.
+            # nb: this is mega huge hacky. THINK. RBC 2006062
+            nested = ui.ui_factory.nested_progress_bar()
+            try:
+                if parents:
+                    self.target.repository.fetch(self.source.repository,
+                        revision_id=parents[0])
+                if threads:
+                    for thread, rev_id in reversed(threads):
+                        # fetch the loom content for this revision
+                        self.target.repository.fetch(self.source.repository,
+                            revision_id=rev_id)
+            finally:
+                nested.finished()
+        state = loom_state.LoomState()
+        try:
+            require_loom_branch(self.target)
+            if threads:
+                last_rev = threads[-1][1]
+                if last_rev == EMPTY_REVISION:
+                    last_rev = bzrlib.revision.NULL_REVISION
+                self.target.generate_revision_history(last_rev)
+                state.set_parents([loom_tip])
+                state.set_threads(
+                    (thread + ([thread[1]],) for thread in threads)
+                    )
+            else:
+                # no threads yet, be a normal branch.
+                self.source._synchronize_history(self.target, revision_id)
+            target_loom = self.unwrap_branch(self.target)
+            target_loom._set_last_loom(state)
+        except NotALoom:
+            self.source._synchronize_history(self.target, revision_id)
+        try:
+            parent = self.source.get_parent()
+        except errors.InaccessibleParent, e:
+            trace.mutter('parent was not accessible to copy: %s', e)
+        else:
+            if parent:
+                self.target.set_parent(parent)
+        if threads:
+            self.target._set_nick(threads[-1][0])
+        if self.source._push_should_merge_tags():
+            self.source.tags.merge_to(self.target.tags)
+
+    @needs_write_lock
+    def pull(self, overwrite=False, stop_revision=None,
+        run_hooks=True, possible_transports=None, _override_hook_target=None,
+        local=False):
+        """Perform a pull, reading from self.source and writing to self.target.
+
+        If the source branch is a non-loom branch, the pull is done against the
+        current warp. If it is a loom branch, then the pull is done against the
+        entire loom and the current thread set to the top thread.
+        """
+        # Special code only needed when both source and targets are looms:
+        if (self.__class__.branch_is_loom(self.target) and
+            self.__class__.branch_is_loom(self.source)):
+            return _Puller(self.source, self.target).transfer(overwrite, stop_revision,
+                run_hooks, possible_transports, _override_hook_target, local)
+        return super(InterLoomBranch, self).pull(
+            overwrite=overwrite, stop_revision=stop_revision,
+            possible_transports=possible_transports,
+            _override_hook_target=_override_hook_target, local=local,
+            run_hooks=run_hooks)
 
 
-LOOM_FORMATS = [
-    BzrBranchLoomFormat1,
-    BzrBranchLoomFormat6,
-    BzrBranchLoomFormat7,
-]
+bzrlib.branch.InterBranch.register_optimiser(InterLoomBranch)
