@@ -35,7 +35,7 @@ class TestsWithLooms(TestCaseWithLoom):
     def _add_patch(self, tree, name):
         """Add a patch to a new thread, returning the revid of te commit."""
         tree.branch.new_thread(name)
-        tree.branch.nick = name
+        tree.branch._set_nick(name)
         self.build_tree([name])
         tree.add(name)
         return tree.commit(name)
@@ -52,7 +52,7 @@ class TestsWithLooms(TestCaseWithLoom):
         tree.commit('first release')
         self.run_bzr(['loomify', path])
         return tree.bzrdir.open_workingtree()
-    
+
     def assert_exception_raised_on_non_loom_branch(self, args):
         """Helper to check UserError gets raised when commands are run in a non-loomed branch."""
         tree = self.make_branch_and_tree('.')
@@ -104,7 +104,7 @@ class TestLoomify(TestCaseWithLoom):
 
 
 class TestCreate(TestsWithLooms):
-    
+
     def test_create_no_changes(self):
         tree = self.get_vendor_loom()
         out, err = self.run_bzr(['create-thread', 'debian'])
@@ -140,18 +140,18 @@ class TestCreate(TestsWithLooms):
 
 
 class TestShow(TestsWithLooms):
-    
+
     def test_show_loom(self):
         """Show the threads in the loom."""
         tree = self.get_vendor_loom()
         self.assertShowLoom(['vendor'], 'vendor')
         tree.branch.new_thread('debian')
         self.assertShowLoom(['vendor', 'debian'], 'vendor')
-        tree.branch.nick = 'debian'
+        tree.branch._set_nick('debian')
         self.assertShowLoom(['vendor', 'debian'], 'debian')
         tree.branch.new_thread('patch A', 'vendor')
         self.assertShowLoom(['vendor', 'patch A', 'debian'], 'debian')
-        tree.branch.nick = 'patch A'
+        tree.branch._set_nick('patch A')
         self.assertShowLoom(['vendor', 'patch A', 'debian'], 'patch A')
 
     def test_show_loom_with_location(self):
@@ -184,6 +184,19 @@ class TestShow(TestsWithLooms):
 
 class TestStatus(TestsWithLooms):
 
+    def setUp(self):
+        super(TestStatus, self).setUp()
+        # The test suite resets after each run, so manually register
+        # the loom status hook.
+        try:
+            from bzrlib.hooks import install_lazy_named_hook
+        except ImportError:
+            pass
+        else:
+            from bzrlib.plugins.loom import show_loom_summary
+            install_lazy_named_hook('bzrlib.status', 'hooks', 'post_status',
+                show_loom_summary, 'loom status')
+
     def test_status_shows_current_thread(self):
         # 'bzr status' shows the current thread.
         tree = self.get_vendor_loom()
@@ -227,7 +240,7 @@ class TestStatus(TestsWithLooms):
 
 
 class TestSwitch(TestsWithLooms):
-    
+
     def test_switch_thread_up_does_not_merge(self):
         tree = self.get_vendor_loom()
         self._add_patch(tree, 'thread1')
@@ -292,7 +305,7 @@ class TestRecord(TestsWithLooms):
         """Adding a new thread is enough to allow recording."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('feature')
-        tree.branch.nick = 'feature'
+        tree.branch._set_nick('feature')
         out, err = self.run_bzr(['record', 'add feature branch.'])
         self.assertEqual('Loom recorded.\n', out)
         self.assertEqual('', err)
@@ -314,7 +327,7 @@ class TestDown(TestsWithLooms):
         """moving down when the revision is unchanged should work."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'patch'
+        tree.branch._set_nick('patch')
         rev = tree.last_revision()
         out, err = self.run_bzr(['down-thread'])
         self.assertEqual('', out)
@@ -325,7 +338,7 @@ class TestDown(TestsWithLooms):
     def test_down_thread_removes_changes_between_threads(self):
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'patch'
+        tree.branch._set_nick('patch')
         rev = tree.last_revision()
         self.build_tree(['afile'])
         tree.add('afile')
@@ -347,7 +360,7 @@ class TestDown(TestsWithLooms):
         """Do a down thread when the lower patch is not in the r-h of the old."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'vendor'
+        tree.branch._set_nick('vendor')
         # do a null change in vendor - a new release.
         vendor_release = tree.commit('new vendor release.', allow_pointless=True)
         # pop up, then down
@@ -400,7 +413,7 @@ class TestDown(TestsWithLooms):
         """Trying to down-thread with changes causes an error."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('upper-thread')
-        tree.branch.nick = 'upper-thread'
+        tree.branch._set_nick('upper-thread')
         self.build_tree(['new-file'])
         tree.add('new-file')
         out, err = self.run_bzr('down-thread', retcode=3)
@@ -416,29 +429,29 @@ class TestUp(TestsWithLooms):
         self.assertEqual('', out)
         self.assertEqual(
             'bzr: ERROR: Cannot move up from the highest thread.\n', err)
-        
+
     def test_up_thread_same_revision(self):
         """moving up when the revision is unchanged should work."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'vendor'
+        tree.branch._set_nick('vendor')
         rev = tree.last_revision()
         out, err = self.run_bzr(['up-thread'])
         self.assertEqual('', out)
         self.assertEqual('', err)
         self.assertEqual('patch', tree.branch.nick)
         self.assertEqual(rev, tree.last_revision())
-        
-    def test_up_thread_preserves_changes(self):
+
+    def test_up_thread_manual_preserves_changes(self):
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'vendor'
+        tree.branch._set_nick('vendor')
         patch_rev = tree.last_revision()
         # add a change in vendor - a new release.
         self.build_tree(['afile'])
         tree.add('afile')
         vendor_release = tree.commit('new vendor release adds a file.')
-        out, err = self.run_bzr(['up-thread'])
+        out, err = self.run_bzr(['up-thread', '--manual'])
         self.assertEqual('', out)
         self.assertEqual(
             "All changes applied successfully.\n"
@@ -455,11 +468,18 @@ class TestUp(TestsWithLooms):
         self.run_bzr(['diff'], retcode=1)
         self.assertEqual([patch_rev, vendor_release], tree.get_parent_ids())
 
+    def test_up_thread_manual_rejects_specified_thread(self):
+        tree = self.get_vendor_loom()
+        tree.branch.new_thread('patch')
+        out, err = self.run_bzr('up-thread --manual patch', retcode=3)
+        self.assertContainsRe(err, 'Specifying a thread does not work with'
+                              ' --manual.')
+
     def test_up_thread_gets_conflicts(self):
         """Do a change in both the baseline and the next patch up."""
         tree = self.get_vendor_loom()
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'patch'
+        tree.branch._set_nick('patch')
         # add a change in patch - a new release.
         self.build_tree(['afile'])
         tree.add('afile')
@@ -496,11 +516,11 @@ class TestUp(TestsWithLooms):
         self.run_bzr(['down-thread'])
         self.run_bzr(['up-thread', '--lca'])
 
-    def test_up_thread_auto(self):
+    def test_up_thread_no_manual(self):
         tree = self.get_vendor_loom()
         tree.branch.new_thread('middle')
         tree.branch.new_thread('top')
-        self.run_bzr('up-thread --auto')
+        self.run_bzr('up-thread')
         branch = _mod_branch.Branch.open('.')
         self.assertEqual('top', branch.nick)
 
@@ -518,7 +538,7 @@ class TestUp(TestsWithLooms):
         tree.add('afile')
         patch_rev = tree.commit('add afile in base')
         tree.branch.new_thread('patch')
-        tree.branch.nick = 'patch'
+        tree.branch._set_nick('patch')
         # make a change to afile in patch.
         f = open('afile', 'wb')
         try:
@@ -536,12 +556,13 @@ class TestUp(TestsWithLooms):
         vendor_release = tree.commit('make the same change to afile')
         # check that the trees no longer differ after the up merge,
         # and that we are 
-        out, err = self.run_bzr(['up-thread'])
+        out, err = self.run_bzr(['up-thread', '--manual'])
         self.assertEqual('', out)
-        self.assertEqual("All changes applied successfully.\n"
+        self.assertStartsWith(err,
+            "All changes applied successfully.\n"
             "Moved to thread 'patch'.\n"
             'This thread is now empty, you may wish to run "bzr '
-            'combine-thread" to remove it.\n', err)
+            'combine-thread" to remove it.\n')
         self.assertEqual('patch', tree.branch.nick)
         # the tree needs to be updated.
         self.assertEqual(patch_rev, tree.last_revision())
@@ -552,6 +573,14 @@ class TestUp(TestsWithLooms):
         self.run_bzr(['diff'])
         self.assertEqual([patch_rev, vendor_release], tree.get_parent_ids())
 
+    def test_up_thread_accepts_thread(self):
+        tree = self.get_vendor_loom()
+        tree.branch.new_thread('lower-middle')
+        tree.branch.new_thread('upper-middle')
+        tree.branch.new_thread('top')
+        self.run_bzr('up-thread upper-middle')
+        branch = _mod_branch.Branch.open('.')
+        self.assertEqual('upper-middle', branch.nick)
 
 
 class TestPush(TestsWithLooms):
@@ -580,7 +609,9 @@ class TestBranch(TestsWithLooms):
         tree.branch.record_loom('commit loom.')
         out, err = self.run_bzr(['branch', 'source', 'target'])
         self.assertEqual('', out)
-        self.assertEqual('Branched 1 revision(s).\n', err)
+        self.assertTrue(
+            err == 'Branched 1 revision(s).\n' or
+            err == 'Branched 1 revision.\n')
         # lower level tests check behaviours, just check show-loom as a smoke
         # test.
         out, err = self.run_bzr(['show-loom', 'target'])
@@ -654,7 +685,7 @@ class TestRevert(TestsWithLooms):
         # not in, so we can revert that by name,
         tree = self.get_vendor_loom()
         tree.branch.new_thread('after-vendor')
-        tree.branch.nick = 'after-vendor'
+        tree.branch._set_nick('after-vendor')
         tree.commit('after-vendor commit', allow_pointless=True)
         tree.branch.record_loom('save loom with vendor and after-vendor')
         old_threads = tree.branch.get_loom_state().get_threads()
@@ -694,6 +725,56 @@ class TestCombineThread(TestsWithLooms):
         loom_tree.down_thread()
         return tree, loom_tree
 
+    def get_loom_with_unique_thread(self):
+        """Return a loom with a unique thread.
+
+        That is:
+        vendor:[]
+        unique-thread:[vendor]
+        above-vendor:[vendor]
+
+        - unique-thread has work not in vendor and not in above-vendor.
+
+        The returned loom is on the vendor thread.
+        """
+        tree, _ = self.get_two_thread_loom()
+        tree.branch.new_thread('unique-thread', 'vendor')
+        loom_tree = LoomTreeDecorator(tree)
+        loom_tree.up_thread()
+        self.build_tree(['file-b'])
+        tree.add('file-b')
+        tree.commit('a unique change', rev_id='uniquely-yrs-1')
+        loom_tree.down_thread()
+        return tree, loom_tree
+
+    def test_combine_unmerged_thread_force(self):
+        """Combining a thread with unique work works with --force."""
+        tree, loom_tree = self.get_loom_with_unique_thread()
+        vendor_revid = tree.last_revision()
+        loom_tree.up_thread()
+        out, err = self.run_bzr(['combine-thread', '--force'])
+        self.assertEqual('', out)
+        self.assertEqual(
+            "Combining thread 'unique-thread' into 'vendor'\n"
+            'All changes applied successfully.\n'
+            "Moved to thread 'vendor'.\n",
+            err)
+        self.assertEqual(vendor_revid, tree.last_revision())
+        self.assertEqual('vendor', tree.branch.nick)
+
+    def test_combine_unmerged_thread_errors(self):
+        """Combining a thread with unique work errors without --force."""
+        tree, loom_tree = self.get_loom_with_unique_thread()
+        loom_tree.up_thread()
+        unique_revid = tree.last_revision()
+        out, err = self.run_bzr(['combine-thread'], retcode=3)
+        self.assertEqual('', out)
+        self.assertEqual("bzr: ERROR: "
+"Thread 'unique-thread' has unmerged work. Use --force to combine anyway.\n",
+            err)
+        self.assertEqual(unique_revid, tree.last_revision())
+        self.assertEqual('unique-thread', tree.branch.nick)
+
     def test_combine_last_two_threads(self):
         """Doing a combine on two threads gives you just the bottom one."""
         tree, loom_tree = self.get_two_thread_loom()
