@@ -23,6 +23,7 @@ import bzrlib
 from bzrlib.branch import Branch
 import bzrlib.errors as errors
 from bzrlib.plugins.loom.branch import (
+    AlreadyLoom,
     EMPTY_REVISION,
     loomify,
     require_loom_branch,
@@ -33,7 +34,10 @@ from bzrlib.plugins.loom.tests import TestCaseWithLoom
 from bzrlib.plugins.loom.tree import LoomTreeDecorator
 import bzrlib.revision
 from bzrlib.revision import NULL_REVISION
-from bzrlib.tests import TestCaseWithTransport
+from bzrlib.tests import (
+    TestCaseWithTransport,
+    test_server,
+    )
 from bzrlib.transport import get_transport
 from bzrlib.workingtree import WorkingTree
 
@@ -48,11 +52,18 @@ class TestFormat(TestCaseWithTransport):
         self.assertFileEqual('Loom current 1\n\n', '.bzr/branch/last-loom')
 
 
+
+class StubFormat(object):
+
+    def network_name(self):
+        return "Nothing to see."
+
+
 class LockableStub(object):
 
     def __init__(self):
         self._calls = []
-        self._format = "Nothing to see."
+        self._format = StubFormat()
 
     def lock_write(self):
         self._calls.append(("write",))
@@ -67,8 +78,7 @@ class TestRequireLoomBranch(TestCaseWithTransport):
         branch = self.make_branch('.')
         self.assertRaises(NotALoom, require_loom_branch, branch)
 
-
-    def works_on_loom(self, format):
+    def works_on_format(self, format):
         branch = self.make_branch('.', format)
         loomify(branch)
         # reopen it
@@ -76,13 +86,19 @@ class TestRequireLoomBranch(TestCaseWithTransport):
         self.assertEqual(None, require_loom_branch(branch))
 
     def test_works_on_loom1(self):
-        self.works_on_loom('knit')
+        self.works_on_format('knit')
 
     def test_works_on_loom6(self):
-        self.works_on_loom('pack-0.92')
+        self.works_on_format('pack-0.92')
 
     def test_works_on_loom7(self):
-        self.works_on_loom('1.6')
+        self.works_on_format('1.6')
+
+    def test_no_harm_to_looms(self):
+        branch = self.make_branch('.')
+        loomify(branch)
+        branch = branch.bzrdir.open_branch()
+        self.assertRaises(AlreadyLoom, loomify, branch)
 
 
 class TestLoomify(TestCaseWithTransport):
@@ -185,9 +201,9 @@ class TestLoom(TestCaseWithLoom):
         tree.branch.new_thread('baseline')
         tree.branch.new_thread('middlepoint')
         tree.branch.new_thread('endpoint')
-        tree.branch.nick = 'middlepoint'
+        tree.branch._set_nick('middlepoint')
         rev_id2 = tree.commit('middle', allow_pointless=True)
-        tree.branch.nick = 'endpoint'
+        tree.branch._set_nick('endpoint')
         rev_id3 = tree.commit('end', allow_pointless=True)
         tree.branch.new_thread('afterbase', 'baseline')
         tree.branch.new_thread('aftermiddle', 'middlepoint')
@@ -210,7 +226,7 @@ class TestLoom(TestCaseWithLoom):
         tree = self.get_tree_with_one_commit()
         tree.branch.new_thread('baseline')
         tree.branch.new_thread('tail')
-        tree.branch.nick = 'baseline'
+        tree.branch._set_nick('baseline')
         first_rev = tree.last_revision()
         # lock the tree to prevent unlock triggering implicit record
         tree.lock_write()
@@ -231,7 +247,7 @@ class TestLoom(TestCaseWithLoom):
 
     def test_clone_empty_loom(self):
         source_tree = self.get_tree_with_loom('source')
-        source_tree.branch.nick = 'source'
+        source_tree.branch._set_nick('source')
         target_tree = source_tree.bzrdir.clone('target').open_workingtree()
         self.assertLoomSproutedOk(source_tree, target_tree)
 
@@ -245,7 +261,7 @@ class TestLoom(TestCaseWithLoom):
         source_tree = self.get_tree_with_one_commit('source')
         source_tree.branch.new_thread('bottom')
         source_tree.branch.new_thread('top')
-        source_tree.branch.nick = 'top'
+        source_tree.branch._set_nick('top')
         source_tree.commit('phwoar', allow_pointless=True)
         source_tree.branch.record_loom('commit to loom')
         target_tree = source_tree.bzrdir.clone('target').open_workingtree()
@@ -253,23 +269,33 @@ class TestLoom(TestCaseWithLoom):
 
     def test_clone_nonempty_loom_bottom(self):
         """Cloning loom should reset the current loom pointer."""
+        self.make_and_clone_simple_loom()
+
+    def make_and_clone_simple_loom(self):
         source_tree = self.get_tree_with_one_commit('source')
         source_tree.branch.new_thread('bottom')
         source_tree.branch.new_thread('top')
-        source_tree.branch.nick = 'top'
+        source_tree.branch._set_nick('top')
         source_tree.commit('phwoar', allow_pointless=True)
         source_tree.branch.record_loom('commit to loom')
         LoomTreeDecorator(source_tree).down_thread()
-        # now clone
-        target_tree = source_tree.bzrdir.clone('target').open_workingtree()
+        # now clone from the 'default url' - transport_server rather than
+        # vfs_server.
+        source_branch = Branch.open(self.get_url('source'))
+        target_tree = source_branch.bzrdir.sprout('target').open_workingtree()
         self.assertLoomSproutedOk(source_tree, target_tree)
 
+    def test_sprout_remote_loom(self):
+        # RemoteBranch should permit sprouting properly.
+        self.transport_server = test_server.SmartTCPServer_for_testing
+        self.make_and_clone_simple_loom()
+
     def test_sprout_nonempty_loom_bottom(self):
         """Sprouting always resets the loom to the top."""
         source_tree = self.get_tree_with_one_commit('source')
         source_tree.branch.new_thread('bottom')
         source_tree.branch.new_thread('top')
-        source_tree.branch.nick = 'top'
+        source_tree.branch._set_nick('top')
         source_tree.commit('phwoar', allow_pointless=True)
         source_tree.branch.record_loom('commit to loom')
         LoomTreeDecorator(source_tree).down_thread()
@@ -312,21 +338,18 @@ class TestLoom(TestCaseWithLoom):
         if source_threads:
             source_tree.branch.generate_revision_history(source_threads[-1][1])
         self.assertEqual(
-            source_tree.branch.revision_history(),
-            target_tree.branch.revision_history())
-        self.assertEqual(
-            source_tree.branch.last_revision(),
-            target_tree.last_revision())
+            source_tree.branch.last_revision_info(),
+            target_tree.branch.last_revision_info())
 
     def test_pull_loom_at_bottom(self):
         """Pulling from a loom when in the bottom warp pulls all warps."""
         source = self.get_tree_with_loom('source')
         source.branch.new_thread('bottom')
         source.branch.new_thread('top')
-        source.branch.nick = 'bottom'
+        source.branch._set_nick('bottom')
         source.branch.record_loom('commit to loom')
         target = source.bzrdir.sprout('target').open_branch()
-        target.nick = 'top'
+        target._set_nick('top')
         # put a commit in the bottom and top of this loom
         bottom_rev1 = source.commit('commit my arse')
         source_loom_tree = LoomTreeDecorator(source)
@@ -356,14 +379,20 @@ class TestLoom(TestCaseWithLoom):
         
     def test_pull_into_empty_loom(self):
         """Doing a pull into a loom with no loom revisions works."""
+        self.pull_into_empty_loom()
+
+    def pull_into_empty_loom(self):
         source = self.get_tree_with_loom('source')
         target = source.bzrdir.sprout('target').open_branch()
         source.branch.new_thread('a thread')
-        source.branch.nick = 'a thread'
+        source.branch._set_nick('a thread')
         # put a commit in the thread for source.
         bottom_rev1 = source.commit('commit a thread')
         source.branch.record_loom('commit to loom')
-        target.pull(source.branch)
+        # now pull from the 'default url' - transport_server rather than
+        # vfs_server - this may be a RemoteBranch.
+        source_branch = Branch.open(self.get_url('source'))
+        target.pull(source_branch)
         # check loom threads
         threads = target.get_loom_state().get_threads()
         self.assertEqual(
@@ -375,12 +404,17 @@ class TestLoom(TestCaseWithLoom):
             self.assertTrue(target.repository.has_revision(rev_id))
         self.assertEqual(source.branch.loom_parents(), target.loom_parents())
 
+    def test_pull_remote_loom(self):
+        # RemoteBranch should permit sprouting properly.
+        self.transport_server = test_server.SmartTCPServer_for_testing
+        self.pull_into_empty_loom()
+
     def test_pull_thread_at_null(self):
         """Doing a pull when the source loom has a thread with no history."""
         source = self.get_tree_with_loom('source')
         target = source.bzrdir.sprout('target').open_branch()
         source.branch.new_thread('a thread')
-        source.branch.nick = 'a thread'
+        source.branch._set_nick('a thread')
         source.branch.record_loom('commit to loom')
         target.pull(source.branch)
         # check loom threads
@@ -399,10 +433,10 @@ class TestLoom(TestCaseWithLoom):
         source = self.get_tree_with_loom('source')
         source.branch.new_thread('bottom')
         source.branch.new_thread('top')
-        source.branch.nick = 'bottom'
+        source.branch._set_nick('bottom')
         source.branch.record_loom('commit to loom')
         target = source.bzrdir.sprout('target').open_branch()
-        target.nick = 'top'
+        target._set_nick('top')
         # put a commit in the bottom and top of this loom
         bottom_rev1 = source.commit('commit bottom')
         source_loom_tree = LoomTreeDecorator(source)
@@ -433,7 +467,7 @@ class TestLoom(TestCaseWithLoom):
     def test_implicit_record(self):
         tree = self.get_tree_with_loom('source')
         tree.branch.new_thread('bottom')
-        tree.branch.nick = 'bottom'
+        tree.branch._set_nick('bottom')
         tree.lock_write()
         try:
             bottom_rev1 = tree.commit('commit my arse')
@@ -454,7 +488,7 @@ class TestLoom(TestCaseWithLoom):
         self.assertEqual([], tree.branch.loom_parents())
         # add a thread and record it.
         tree.branch.new_thread('bottom')
-        tree.branch.nick = 'bottom'
+        tree.branch._set_nick('bottom')
         rev_id = tree.branch.record_loom('Setup test loom.')
         # after recording, the parents list should have changed.
         self.assertEqual([rev_id], tree.branch.loom_parents())
@@ -465,7 +499,7 @@ class TestLoom(TestCaseWithLoom):
         # new threads
         tree.branch.new_thread('foo')
         tree.branch.new_thread('bar')
-        tree.branch.nick = 'bar'
+        tree.branch._set_nick('bar')
         last_rev = tree.branch.last_revision()
         # and a change to the revision history of this thread
         tree.commit('change bar', allow_pointless=True)
@@ -479,7 +513,7 @@ class TestLoom(TestCaseWithLoom):
         # new threads
         tree.branch.new_thread('foo')
         tree.branch.new_thread('bar')
-        tree.branch.nick = 'bar'
+        tree.branch._set_nick('bar')
         # and a change to the revision history of this thread
         tree.commit('change bar', allow_pointless=True)
         # now record
@@ -504,7 +538,7 @@ class TestLoom(TestCaseWithLoom):
         # new threads
         tree.branch.new_thread('base')
         tree.branch.new_thread('top')
-        tree.branch.nick = 'top'
+        tree.branch._set_nick('top')
         # and a change to the revision history of this thread
         tree.tree.commit('change top', allow_pointless=True)
         last_rev = tree.branch.last_revision()
@@ -528,7 +562,7 @@ class TestLoom(TestCaseWithLoom):
         tree.branch.new_thread('foo')
         tree.branch.new_thread('bar')
         # do a commit, so the last_revision should change.
-        tree.branch.nick = 'bar'
+        tree.branch._set_nick('bar')
         tree.commit('bar-ness', allow_pointless=True)
         tree.branch.revert_thread('bar')
         self.assertEqual(
@@ -541,11 +575,11 @@ class TestLoom(TestCaseWithLoom):
         # ensure we have some stuff to revert
         tree.branch.new_thread('foo')
         tree.branch.new_thread('bar')
-        tree.branch.nick = 'foo'
+        tree.branch._set_nick('foo')
         # record the loom to put the threads in the basis
         tree.branch.record_loom('record it!')
         # do a commit, so the last_revision should change.
-        tree.branch.nick = 'bar'
+        tree.branch._set_nick('bar')
         tree.commit('bar-ness', allow_pointless=True)
         tree.branch.revert_thread('bar')
         self.assertEqual(
@@ -559,7 +593,7 @@ class TestLoom(TestCaseWithLoom):
         tree = self.get_tree_with_loom()
         tree.branch.new_thread('bar')
         tree.branch.new_thread('foo')
-        tree.branch.nick = 'bar'
+        tree.branch._set_nick('bar')
         tree.branch.remove_thread('foo')
         state = tree.branch.get_loom_state()
         self.assertEqual([('bar', 'empty:', [])], state.get_threads())
@@ -570,17 +604,17 @@ class TestLoom(TestCaseWithLoom):
         self.assertEqual([], tree.branch.get_threads(NULL_REVISION))
         # and loom history should make no difference:
         tree.branch.new_thread('foo')
-        tree.branch.nick = 'foo'
+        tree.branch._set_nick('foo')
         tree.branch.record_loom('foo')
         self.assertEqual([], tree.branch.get_threads(NULL_REVISION))
 
     def get_multi_threaded(self):
         tree = self.get_tree_with_loom()
         tree.branch.new_thread('thread1')
-        tree.branch.nick = 'thread1'
+        tree.branch._set_nick('thread1')
         tree.commit('thread1', rev_id='thread1-id')
         tree.branch.new_thread('thread2', 'thread1')
-        tree.branch.nick = 'thread2'
+        tree.branch._set_nick('thread2')
         tree.commit('thread2', rev_id='thread2-id')
         return tree
 
@@ -636,3 +670,14 @@ class TestLoom(TestCaseWithLoom):
         export_branch = Branch.open_from_transport(
             root_transport.clone('thread1'))
         self.assertEqual('thread1-id', export_branch.last_revision())
+
+    def test_set_nick_renames_thread(self):
+        tree = self.get_tree_with_loom()
+        tree.branch.new_thread(tree.branch.nick)
+        orig_threads = tree.branch.get_loom_state().get_threads()
+        new_thread_name = 'new thread name'
+        tree.branch.nick = new_thread_name
+        new_threads = tree.branch.get_loom_state().get_threads()
+        self.assertNotEqual(orig_threads, new_threads)
+        self.assertEqual(new_thread_name, new_threads[0][0])
+        self.assertEqual(new_thread_name, tree.branch.nick)
